//! File system in os
mod inode;
mod stdio;
mod pipe;
mod stat;

use crate::mm::UserBuffer;

#[derive(Clone)]
/// 文件描述符结构体
pub struct FileDescripter {
    /// close-on-exec 标志，表示文件描述符在执行新程序时是否自动关闭
    cloexec: bool,
    /// 文件类别
    pub fclass: FileClass,
}

impl FileDescripter {
    /// 创建一个新的文件描述符对象。
    pub fn new(cloexec:bool, fclass:FileClass)->Self{
        Self{
            cloexec,
            fclass
        }
    }
    /// 设置 close-on-exec 标志。
    pub fn set_cloexec(&mut self, flag: bool){
        self.cloexec = flag;
    }
    /// 获取当前的 close-on-exec 标志值。
    pub fn get_cloexec(& self) -> bool{
        self.cloexec
    }
}

#[derive(Clone)]
/// 表示文件的类别。
pub enum FileClass {
    /// 文件类型，包含一个指向操作系统节点的强引用。
    File(Arc<OSInode>),
    /// 抽象文件类型，包含一个实现了 `File`、`Send` 和 `Sync` 特性的强引用对象。
    Abstr(Arc<dyn File + Send + Sync>),
}

/// File trait
pub trait File: Send + Sync {
    /// If readable
    fn readable(&self) -> bool;
    /// If writable
    fn writable(&self) -> bool;
    /// Read file to `UserBuffer` 
    fn read(&self, buf: UserBuffer) -> usize;                 
     /// Read file to `UserBuffer`
    fn read2(&self, buf: UserBuffer) -> Result<usize, isize>;
    /// Write `UserBuffer` to file
    fn write(&self, buf: UserBuffer) -> usize;
    /// Write `UserBuffer` to file
    fn write2(&self, buf: UserBuffer) -> Result<usize, isize>;
}



pub use inode::{list_apps, open_file, OSInode, OpenFlags,is_abs_path};
pub use stdio::{Stdin, Stdout};
pub use pipe::make_pipe;
pub use stat::KStat;
use alloc::sync::Arc;

