use super::File; // 引入上一级模块中的File类型
use crate::mm::UserBuffer; // 引入当前crate中mm模块下的UserBuffer类型
use crate::task::suspend_current_and_run_next; // 引入当前crate中task模块下的suspend_current_and_run_next函数
use alloc::sync::{Arc, Weak}; // 引入alloc crate中sync模块下的Arc和Weak类型
use spin::Mutex; // 引入spin crate中的Mutex类型

pub struct Pipe {
    readable: bool,
    writable: bool,
    buffer: Arc<Mutex<PipeRingBuffer>>, // buffer是一个Arc类型的指向Mutex类型的PipeRingBuffer的引用
}

impl Pipe {
    pub fn read_end_with_buffer(buffer: Arc<Mutex<PipeRingBuffer>>) -> Self {
        Self {
            readable: true,
            writable: false,
            buffer,
        }
    }
    pub fn write_end_with_buffer(buffer: Arc<Mutex<PipeRingBuffer>>) -> Self {
        Self {
            readable: false,
            writable: true,
            buffer,
        }
    }
}

const RING_BUFFER_SIZE: usize = 32;

#[derive(Copy, Clone, PartialEq)]
enum RingBufferStatus {
    FULL,
    EMPTY,
    NORMAL,
}

pub struct PipeRingBuffer {
    arr: [u8; RING_BUFFER_SIZE], // 长度为RING_BUFFER_SIZE的u8数组
    head: usize, // 缓冲区头指针
    tail: usize, // 缓冲区尾指针
    status: RingBufferStatus, // 缓冲区状态
    write_end: Option<Weak<Pipe>>, // 弱引用指向Pipe类型的write_end
}

impl PipeRingBuffer {
    pub fn new() -> Self {
        Self {
            arr: [0; RING_BUFFER_SIZE], // 初始化数组为0
            head: 0,
            tail: 0,
            status: RingBufferStatus::EMPTY, // 初始化状态为EMPTY
            write_end: None, // 初始化write_end为None
        }
    }
    pub fn set_write_end(&mut self, write_end: &Arc<Pipe>) { // 设置write_end
        self.write_end = Some(Arc::downgrade(write_end)); // 使用Arc::downgrade函数创建一个弱引用并赋给write_end
    }
    pub fn write_byte(&mut self, byte: u8) { // 写入一个字节到缓冲区
        self.status = RingBufferStatus::NORMAL; // 更新状态为NORMAL
        self.arr[self.tail] = byte; // 在尾指针处写入字节
        self.tail = (self.tail + 1) % RING_BUFFER_SIZE; // 更新尾指针位置
        if self.tail == self.head { // 如果尾指针与头指针重合
            self.status = RingBufferStatus::FULL; // 更新状态为FULL
        }
    }
    pub fn read_byte(&mut self) -> u8 { // 从缓冲区中读取一个字节
        self.status = RingBufferStatus::NORMAL; // 更新状态为NORMAL
        let c = self.arr[self.head]; // 读取头指针处的字节
        self.head = (self.head + 1) % RING_BUFFER_SIZE; // 更新头指针位置
        if self.head == self.tail { // 如果头指针与尾指针重合
            self.status = RingBufferStatus::EMPTY; // 更新状态为EMPTY
        }
        c // 返回读取的字节
    }
    pub fn available_read(&self) -> usize { // 返回可读取的字节数
        if self.status == RingBufferStatus::EMPTY { // 如果状态为空
            0 // 返回0
        } else if self.tail > self.head { // 如果尾指针大于头指针
            self.tail - self.head // 返回尾指针减去头指针的值
        } else {
            self.tail + RING_BUFFER_SIZE - self.head // 返回尾指针加上环形缓冲区大小减去头指针的值
        }
    }
    pub fn available_write(&self) -> usize { // 返回可写入的字节数
        if self.status == RingBufferStatus::FULL { // 如果状态为满
            0 // 返回0
        } else {
            RING_BUFFER_SIZE - self.available_read() // 返回环形缓冲区大小减去可读取的字节数
        }
    }
    pub fn all_write_ends_closed(&self) -> bool { // 检查所有的写入端是否都已关闭
        self.write_end.as_ref().unwrap().upgrade().is_none() // 检查所有的写入端是否都已关闭
    }
}

/// Return (read_end, write_end)
pub fn make_pipe() -> (Arc<Pipe>, Arc<Pipe>) { // 创建管道
    let buffer = Arc::new(Mutex::new(PipeRingBuffer::new())); // 创建一个引用计数指向互斥锁的缓冲区
    let read_end = Arc::new(Pipe::read_end_with_buffer(buffer.clone())); // 通过缓冲区创建读取端
    let write_end = Arc::new(Pipe::write_end_with_buffer(buffer.clone())); // 通过缓冲区创建写入端
    buffer.lock().set_write_end(&write_end); // 设置缓冲区的写入端
    (read_end, write_end) // 返回读取端和写入端
}

impl File for Pipe {
    // 实现File trait的read2函数
    fn read2(&self, buf: UserBuffer) -> Result<usize, isize> {
        assert!(self.readable); // 断言可读
        let mut buf_iter = buf.into_iter(); // 转换成迭代器
        let mut read_size = 0usize; // 初始化读取大小为0
        loop {
            let mut ring_buffer = self.buffer.lock(); // 获取缓冲区的互斥锁
            let loop_read = ring_buffer.available_read(); // 可读取的字节数
            if loop_read == 0 { // 如果可读取的字节数为0
                if ring_buffer.all_write_ends_closed() { // 检查所有写入端是否关闭
                    return Ok(read_size); // 如果是，返回读取大小
                }
                drop(ring_buffer); // 释放锁
                suspend_current_and_run_next(); // 挂起当前任务并运行下一个任务
                continue; // 继续循环
            }
            // 最多读取loop_read字节
            for _ in 0..loop_read {
                if let Some(byte_ref) = buf_iter.next() { // 如果存在下一个字节的引用
                    unsafe {
                        *byte_ref = ring_buffer.read_byte(); // 将读取的字节写入缓冲区中
                    }
                    read_size += 1; // 读取大小加一
                } else {
                    return Ok(read_size); // 返回读取大小
                }
            }
        }
    }
    // 实现File trait的write2函数
    fn write2(&self, buf: UserBuffer) -> Result<usize, isize> {
        assert!(self.writable); // 断言可写
        let mut buf_iter = buf.into_iter(); // 转换成迭代器
        let mut write_size = 0usize; // 初始化写入大小为0
        loop {
            let mut ring_buffer = self.buffer.lock(); // 获取缓冲区的互斥锁
            let loop_write = ring_buffer.available_write(); // 可写入的字节数
            if loop_write == 0 { // 如果可写入的字节数为0
                drop(ring_buffer); // 释放锁
                suspend_current_and_run_next(); // 挂起当前任务并运行下一个任务
                continue; // 继续循环
            }
            // 最多写入loop_write字节
            for _ in 0..loop_write {
                if let Some(byte_ref) = buf_iter.next() { // 如果存在下一个字节的引用
                    ring_buffer.write_byte(unsafe { *byte_ref }); // 将字节写入缓冲区中
                    write_size += 1; // 写入大小加一
                } else {
                    return Ok(write_size); // 返回写入大小
                }
            }
        }
    }

    // 实现File trait的readable函数
    fn readable(&self) -> bool {
        todo!() 
    }

    // 实现File trait的writable函数
    fn writable(&self) -> bool {
        todo!() 
    }

    // 实现File trait的read函数
    fn read(&self, _buf: UserBuffer) -> usize {
        todo!() 
    }

    // 实现File trait的write函数
    fn write(&self, _buf: UserBuffer) -> usize {
        todo!() 
    }
}