//! Implementation of [`MapArea`] and [`MemorySet`].
use super::{frame_alloc, FrameTracker};
use super::{PTEFlags, PageTable, PageTableEntry};
use super::{PhysAddr, PhysPageNum, VirtAddr, VirtPageNum};
use super::{StepByOne, VPNRange};
use crate::config::{MEMORY_END, MMIO, PAGE_SIZE, TRAMPOLINE, TRAP_CONTEXT, USER_STACK_SIZE, THREAD_STACK_SIZE_MIN, USER_SIGNAL_STACK, SIGNAL_STACK_SIZE};
use crate::sync::UPSafeCell;
use alloc::collections::BTreeMap;
use alloc::sync::Arc;
use alloc::vec::Vec;
use core::arch::asm;
use lazy_static::*;
use riscv::register::satp;

extern "C" {
    fn stext();
    fn etext();
    fn srodata();
    fn erodata();
    fn sdata();
    fn edata();
    fn sbss_with_stack();
    fn ebss();
    fn ekernel();
    fn strampoline();
}

lazy_static! {
    /// a memory set instance through lazy_static! managing kernel space
    pub static ref KERNEL_SPACE: Arc<UPSafeCell<MemorySet>> =
        Arc::new(unsafe { UPSafeCell::new(MemorySet::new_kernel()) });
}
///Get kernelspace root ppn
pub fn kernel_token() -> usize {
    KERNEL_SPACE.exclusive_access().token()
}
/// memory set structure, controls virtual-memory space
#[allow(unused)]
pub struct MemorySet {
    page_table: PageTable,
    areas: Vec<MapArea>,
    chunks: ChunkArea,
    stack_chunks: ChunkArea,
    mmap_chunks: Vec<ChunkArea>,
}

impl MemorySet {
    
    ///Create an empty `MemorySet`
    pub fn new_bare() -> Self {
        Self {
            page_table: PageTable::new(),
            areas: Vec::new(),
            chunks: ChunkArea::new(MapType::Framed,
                MapPermission::R | MapPermission::W | MapPermission::U),
            mmap_chunks: Vec::new(),
            stack_chunks: ChunkArea::new(MapType::Framed,
                MapPermission::R | MapPermission::W | MapPermission::U),
        }
    }
    ///Get pagetable `root_ppn`
    pub fn token(&self) -> usize {
        self.page_table.token()
    }
    /// Assume that no conflicts.
    pub fn insert_framed_area(
        &mut self,
        start_va: VirtAddr,
        end_va: VirtAddr,
        permission: MapPermission,
    ) {
        self.push(
            MapArea::new(start_va, end_va, MapType::Framed, permission),
            None,
        );
    }
    ///Remove `MapArea` that starts with `start_vpn`
    pub fn remove_area_with_start_vpn(&mut self, start_vpn: VirtPageNum) {
        if let Some((idx, area)) = self
            .areas
            .iter_mut()
            .enumerate()
            .find(|(_, area)| area.vpn_range.get_start() == start_vpn)
        {
            area.unmap(&mut self.page_table);
            self.areas.remove(idx);
        }
    }
    fn push(&mut self, mut map_area: MapArea, data: Option<&[u8]>) {
        map_area.map(&mut self.page_table);
        if let Some(data) = data {
            map_area.copy_data(&mut self.page_table, data, 0);
        }
        self.areas.push(map_area);
    }
    fn push_with_offset(&mut self, mut map_area: MapArea, offset: usize, data: Option<&[u8]>){
        // println!{"3"}
        map_area.map(&mut self.page_table);
        if let Some(data) = data {
            map_area.copy_data(&mut self.page_table, data, offset);
        }
        self.areas.push(map_area);
    }
    /// Mention that trampoline is not collected by areas.
    fn map_trampoline(&mut self) {
        self.page_table.map(
            VirtAddr::from(TRAMPOLINE).into(),
            PhysAddr::from(strampoline as usize).into(),
            PTEFlags::R | PTEFlags::X,
        );
    }
    #[allow(unused)]
    fn map_kernel_shared(&mut self){
        self.page_table.map_kernel_shared();
    }
    ///将内存映射区（Memory Mapping Area）插入到页表中
    pub fn insert_mmap_area(&mut self, start_va: VirtAddr, end_va: VirtAddr, permission: MapPermission) {
        let mut new_chunk_area = ChunkArea::new(MapType::Framed, permission,);
        new_chunk_area.set_mmap_range(start_va, end_va);
        self.mmap_chunks.push(new_chunk_area);

        //self.push_mmap(MapArea::new(
        //    start_va,
        //    (end_va.0).into(),  // bin lazy (X
        //    MapType::Framed,
        //    permission,
        //), None);
    }
    /// Without kernel stacks.
    pub fn new_kernel() -> Self {
        let mut memory_set = Self::new_bare();
        // map trampoline
        memory_set.map_trampoline();
        // map kernel sections
        println!(".text [{:#x}, {:#x})", stext as usize, etext as usize);
        println!(".rodata [{:#x}, {:#x})", srodata as usize, erodata as usize);
        println!(".data [{:#x}, {:#x})", sdata as usize, edata as usize);
        println!(
            ".bss [{:#x}, {:#x})",
            sbss_with_stack as usize, ebss as usize
        );
        println!("mapping .text section");
        memory_set.push(
            MapArea::new(
                (stext as usize).into(),
                (etext as usize).into(),
                MapType::Identical,
                MapPermission::R | MapPermission::X,
            ),
            None,
        );
        println!("mapping .rodata section");
        memory_set.push(
            MapArea::new(
                (srodata as usize).into(),
                (erodata as usize).into(),
                MapType::Identical,
                MapPermission::R,
            ),
            None,
        );
        println!("mapping .data section");
        memory_set.push(
            MapArea::new(
                (sdata as usize).into(),
                (edata as usize).into(),
                MapType::Identical,
                MapPermission::R | MapPermission::W,
            ),
            None,
        );
        println!("mapping .bss section");
        memory_set.push(
            MapArea::new(
                (sbss_with_stack as usize).into(),
                (ebss as usize).into(),
                MapType::Identical,
                MapPermission::R | MapPermission::W,
            ),
            None,
        );
        println!("mapping physical memory");
        memory_set.push(
            MapArea::new(
                (ekernel as usize).into(),
                MEMORY_END.into(),
                MapType::Identical,
                MapPermission::R | MapPermission::W,
            ),
            None,
        );
        println!("mapping memory-mapped registers");
        for pair in MMIO {
            memory_set.push(
                MapArea::new(
                    (*pair).0.into(),
                    ((*pair).0 + (*pair).1).into(),
                    MapType::Identical,
                    MapPermission::R | MapPermission::W,
                ),
                None,
            );
        }
        memory_set
    }
    /// is_mapped_area 判断给定的虚拟地址区间是否已经映射到物理内存
    fn is_mapped_area(&self, start_va: VirtAddr, end_va: VirtAddr) -> bool {
        // 遍历当前进程中的所有内存区域
        for area in &self.areas {
            // 检查当前内存区域的虚拟页号范围是否与指定的地址区间重叠
            if area
                .vpn_range
                .is_overlapped(&VPNRange::new(start_va.into(), end_va.into()))
            {
                // 如果重叠，表示至少有一个页面已经映射到物理内存，返回 true
                return true;
            }
        }
        // 遍历完所有内存区域，都没有找到与指定的地址区间重叠的内存区域，返回 false
        false
    }
    ///mmap 执行映射操作
    pub fn mmap(&mut self, start: usize, len: usize, port: usize) -> Result<isize, isize> {
        // 判断权限端口是否符合要求，如果不符合返回执行映射操作失败的错误结果
        if port & !7 != 0 || port & 7 == 0 || len > 1 << 30 {
            return Err(-1);
        } else {
            let start_va: VirtAddr = VirtAddr::from(start);
            // 判断起始地址是否是虚拟地址的整数倍，如果不是返回执行映射操作失败的错误结果
            if start_va != start_va.floor().into() {
                return Err(-1);
            }
            let end_va: VirtAddr = VirtAddr::from(start + len).ceil().into();

            // 检查起始地址和结束地址之间是否有其他映射区域，如果有返回执行映射操作失败的错误结果
            if self.is_mapped_area(start_va, end_va) {
                return Err(-1);
            }
            // 将映射区域插入到映射表中
            self.insert_framed_area(
                start_va,
                end_va,
                MapPermission::from_bits((port << 1 | 0b10000) as u8).unwrap(),
            );

            // 返回映射的大小（以字节为单位）
            Ok((usize::from(end_va) - usize::from(start_va)) as isize)
        }
    }
    /// munmap 解除内存映射
    pub fn munmap(&mut self, start: usize, len: usize) -> Result<isize, isize> {
        let mut start_va: VirtAddr = VirtAddr::from(start);// 将 start 转换为 VirtAddr，表示虚拟地址
        if start_va != start_va.floor().into() {// 检查转换后的地址与原始地址是否一致
            // 如果地址不是页对齐的，则返回错误
            return Err(-1);
        }
        let end_va: VirtAddr = VirtAddr::from(start + len).ceil().into();// 计算要解除映射的结束地址
        let mut to_unmap: Vec<usize> = Vec::new();// 创建一个空的向量来存储要解除映射的内存区域
        // 遍历所有地址区间，判断是否与要解除映射的地址区间重叠
        // 如果重叠，则将其索引添加到 to_unmap 向量中
        for (i, area) in self.areas.iter().enumerate() {
            if area
                .vpn_range
                .is_overlapped(&VPNRange::new(start_va.into(), end_va.into()))
            {
                to_unmap.push(i);
            }
        }
    
        // 按照地址区间的起始地址排序
        to_unmap.sort_by_key(|i| self.areas[*i].vpn_range.get_start());
    
        // 遍历所有索引，检查游标地址是否与下一内存区间的起始地址相等
        // 如果相等，则将游标地址更新为下一内存区间的结束地址
        // 否则，返回错误
        for i in &to_unmap {
            if start_va == self.areas[*i].vpn_range.get_start().into() {
                start_va = self.areas[*i].vpn_range.get_end().into();
            } else {
                return Err(-1);
            }
        }
        // 检查游标地址是否与结束地址相等
        // 如果不相等，则返回错误
        if start_va != end_va {
            return Err(-1);
        }
    
        // 按照要解除映射的内存区间的索引从大到小进行排序
        to_unmap.sort_by(|l, r| r.cmp(l));
    
        // 遍历所有要解除映射的内存区间的索引
        // 解除映射并从内存区域列表中移除相应的内存区域
        for i in to_unmap {
            self.areas[i].unmap(&mut self.page_table);
            self.areas.remove(i);
        }
    
        // 返回解除映射的长度
        Ok(len as isize)
    }
    /// Include sections in elf and trampoline and TrapContext and user stack,
    /// also returns user_sp and entry point.


    pub fn from_elf(elf_data: &[u8]) -> (Self, usize, usize, usize) {
        let mut memory_set = Self::new_bare();
        // map trampoline
        memory_set.map_trampoline();
        // map program headers of elf, with U flag
        let elf = xmas_elf::ElfFile::new(elf_data).unwrap();
        let elf_header = elf.header;
        let magic = elf_header.pt1.magic;
        assert_eq!(magic, [0x7f, 0x45, 0x4c, 0x46], "invalid elf!");
        let ph_count = elf_header.pt2.ph_count();
        let mut max_end_vpn = VirtPageNum(0);
        let mut head_va = 0; // top va of ELF which points to ELF header

        //let mut comment_flag = true;

        for i in 0..ph_count {
            let ph = elf.program_header(i).unwrap();
            if ph.get_type().unwrap() == xmas_elf::program::Type::Load {
                let start_va: VirtAddr = (ph.virtual_addr() as usize).into();
                let end_va: VirtAddr = ((ph.virtual_addr() + ph.mem_size()) as usize).into();
                let offset = start_va.0 - start_va.floor().0 * PAGE_SIZE;

                //if start_va.0 == 0{
                //    comment_flag = false;
                //} 
                

                let mut map_perm = MapPermission::U;
                let ph_flags = ph.flags();
                if ph_flags.is_read() {
                    map_perm |= MapPermission::R;
                }
                if ph_flags.is_write() {
                    map_perm |= MapPermission::W;
                }
                if ph_flags.is_execute() {
                    map_perm |= MapPermission::X;
                }

                let map_area = MapArea::new(start_va, end_va, MapType::Framed, map_perm);
                max_end_vpn = map_area.vpn_range.get_end();
                
                if offset == 0 {
                    head_va = start_va.into();
                    memory_set.push( 
                        map_area,
                        Some(&elf.input[ph.offset() as usize..(ph.offset() + ph.file_size()) as usize])
                    );
                } else {
                    memory_set.push_with_offset( 
                        map_area,
                        offset,
                        Some(&elf.input[ph.offset() as usize..(ph.offset() + ph.file_size()) as usize])
                    );
                }
            }
        }
        
        let _ph_head_addr = head_va + elf.header.pt2.ph_offset() as usize;

        // map user stack with U flags
        let max_end_va: VirtAddr = max_end_vpn.into();
        let mut user_stack_bottom: usize = max_end_va.into();
        // guard page
        user_stack_bottom += PAGE_SIZE;
        let user_stack_top = user_stack_bottom + USER_STACK_SIZE;
        memory_set.push(
            MapArea::new(
                user_stack_bottom.into(),
                user_stack_top.into(),
                MapType::Framed,
                MapPermission::R | MapPermission::W | MapPermission::U,
            ),
            None,
        );
        // map TrapContext
        memory_set.push(
            MapArea::new(
                TRAP_CONTEXT.into(),
                TRAMPOLINE.into(),
                MapType::Framed,
                MapPermission::R | MapPermission::W,
            ),
            None,
        );

        memory_set.map_kernel_shared();
        
        (memory_set, user_stack_top, user_stack_bottom, elf.header.pt2.entry_point() as usize)
    }
    ///Clone a same `MemorySet`
    pub fn from_existed_user(user_space: &MemorySet) -> MemorySet {
        let mut memory_set = Self::new_bare();
        // map trampoline
        memory_set.map_trampoline();
        // copy data sections/trap_context/user_stack
        for area in user_space.areas.iter() {
            let new_area = MapArea::from_another(area);
            memory_set.push(new_area, None);
            // copy data from another space
            for vpn in area.vpn_range {
                let src_ppn = user_space.translate(vpn).unwrap().ppn();
                let dst_ppn = memory_set.translate(vpn).unwrap().ppn();
                dst_ppn
                    .get_bytes_array()
                    .copy_from_slice(src_ppn.get_bytes_array());
            }
        }
        memory_set
    }
    ///Refresh TLB with `sfence.vma`
    pub fn activate(&self) {
        let satp = self.page_table.token();
        unsafe {
            satp::write(satp);
            asm!("sfence.vma");
        }
    }
    ///Translate throuth pagetable
    pub fn translate(&self, vpn: VirtPageNum) -> Option<PageTableEntry> {
        self.page_table.translate(vpn)
    }
    ///设置 VPN 标志位
    pub fn set_pte_flags(&mut self, vpn: VirtPageNum, flags: usize) -> isize{
        self.page_table.set_pte_flags(vpn, flags)
    }
    ///Remove all `MapArea`
    pub fn recycle_data_pages(&mut self) {
        //*self = Self::new_bare();
        self.areas.clear();
    }
    
}
/// map area structure, controls a contiguous piece of virtual memory
pub struct MapArea {
    vpn_range: VPNRange,
    data_frames: BTreeMap<VirtPageNum, FrameTracker>,
    map_type: MapType,
    map_perm: MapPermission,
}
pub struct ChunkArea {
    vpn_table: Vec<VirtPageNum>,
    data_frames: BTreeMap<VirtPageNum, FrameTracker>,
    map_type: MapType,
    map_perm: MapPermission,
    mmap_start: VirtAddr,
    mmap_end: VirtAddr,
}
#[allow(unused)]
impl ChunkArea {
    pub fn new(
        map_type: MapType,
        map_perm: MapPermission
    ) -> Self {
        Self {
            vpn_table: Vec::new(),
            data_frames: BTreeMap::new(),
            map_type,
            map_perm,
            mmap_start: 0.into(),
            mmap_end: 0.into(),
        }
    }

    pub fn set_mmap_range(&mut self, start: VirtAddr, end: VirtAddr) {
        self.mmap_start = start;
        self.mmap_end = end;
    }

    pub fn push_vpn(&mut self, vpn: VirtPageNum, page_table: &mut PageTable) {
        self.vpn_table.push(vpn);
        self.map_one(page_table, vpn);
    }

    pub fn from_another(another: &ChunkArea) -> Self {
        Self {
            vpn_table: another.vpn_table.clone(),
            data_frames: BTreeMap::new(),
            map_type: another.map_type,
            map_perm: another.map_perm,
            mmap_start: another.mmap_start,
            mmap_end: another.mmap_end,
        }
    }

    // Alloc and map one page
    pub fn map_one(&mut self, page_table: &mut PageTable, vpn: VirtPageNum) {
        let ppn: PhysPageNum;
        match self.map_type {
            MapType::Identical => {
                ppn = PhysPageNum(vpn.0);
            }
            MapType::Framed => {
                if let Some(frame) = frame_alloc(){
                    ppn = frame.ppn;
                    self.data_frames.insert(vpn, frame);
                }
                else{
                    panic!("No more memory!");
                }
            }
        }
        let pte_flags = PTEFlags::from_bits(self.map_perm.bits).unwrap();
        // [WARNING]:因为没有map，所以不能使用
        page_table.map(vpn, ppn, pte_flags);
    }
    pub fn unmap_one(&mut self, page_table: &mut PageTable, vpn: VirtPageNum) {
        match self.map_type {
            MapType::Framed => {
                self.data_frames.remove(&vpn);
            }
            _ => {}
        }
        page_table.unmap(vpn);
    }
    
    // Alloc and map all pages
    // pub fn map(&mut self, page_table: &mut PageTable) {
    //     for vpn in self.vpn_table {
    //         self.map_one(page_table, vpn);
    //     }
    // }
    // pub fn unmap(&mut self, page_table: &mut PageTable) {
    //     for vpn in self.vpn_table {
    //         self.unmap_one(page_table, vpn);
    //     }
    // }

}
impl MapArea {
    pub fn new(
        start_va: VirtAddr,
        end_va: VirtAddr,
        map_type: MapType,
        map_perm: MapPermission,
    ) -> Self {
        let start_vpn: VirtPageNum = start_va.floor();
        let end_vpn: VirtPageNum = end_va.ceil();
        Self {
            vpn_range: VPNRange::new(start_vpn, end_vpn),
            data_frames: BTreeMap::new(),
            map_type,
            map_perm,
        }
    }
    pub fn from_another(another: &MapArea) -> Self {
        Self {
            vpn_range: VPNRange::new(another.vpn_range.get_start(), another.vpn_range.get_end()),
            data_frames: BTreeMap::new(),
            map_type: another.map_type,
            map_perm: another.map_perm,
        }
    }
    pub fn map_one(&mut self, page_table: &mut PageTable, vpn: VirtPageNum) {
        let ppn: PhysPageNum;
        match self.map_type {
            MapType::Identical => {
                ppn = PhysPageNum(vpn.0);
            }
            MapType::Framed => {
                let frame = frame_alloc().unwrap();
                ppn = frame.ppn;
                self.data_frames.insert(vpn, frame);
            }
        }
        let pte_flags = PTEFlags::from_bits(self.map_perm.bits).unwrap();
        page_table.map(vpn, ppn, pte_flags);
    }
    pub fn unmap_one(&mut self, page_table: &mut PageTable, vpn: VirtPageNum) {
        if self.map_type == MapType::Framed {
            self.data_frames.remove(&vpn);
        }
        page_table.unmap(vpn);
    }
    pub fn map(&mut self, page_table: &mut PageTable) {
        for vpn in self.vpn_range {
            self.map_one(page_table, vpn);
        }
    }
    pub fn unmap(&mut self, page_table: &mut PageTable) {
        for vpn in self.vpn_range {
            self.unmap_one(page_table, vpn);
        }
    }
    /// data: start-aligned but maybe with shorter length
    /// assume that all frames were cleared before
    pub fn copy_data(&mut self, page_table: &mut PageTable, data: &[u8], offset:usize) {
        assert_eq!(self.map_type, MapType::Framed);
        let mut start: usize = 0;
        let mut page_offset: usize = offset;
        let mut current_vpn = self.vpn_range.get_start();
        let len = data.len();
        loop {
            let src = &data[start..len.min(start + PAGE_SIZE- page_offset)];
            let dst = &mut page_table
                .translate(current_vpn)
                .unwrap()
                .ppn()
                .get_bytes_array()[page_offset..(page_offset+src.len())];
            dst.copy_from_slice(src);

            start += PAGE_SIZE - page_offset;

            page_offset = 0;                                                      
            if start >= len {
                break;
            }
            current_vpn.step();
        }
    }
}

#[derive(Copy, Clone, PartialEq, Debug)]
/// map type for memory set: identical or framed
pub enum MapType {
    Identical,
    Framed,
}

bitflags! {
    /// map permission corresponding to that in pte: `R W X U`
    pub struct MapPermission: u8 {
        ///Readable
        const R = 1 << 1;
        ///Writable
        const W = 1 << 2;
        ///Excutable
        const X = 1 << 3;
        ///Accessible in U mode
        const U = 1 << 4;
    }
}

#[allow(unused)]
///Check PageTable running correctly
pub fn remap_test() {
    let mut kernel_space = KERNEL_SPACE.exclusive_access();
    let mid_text: VirtAddr = ((stext as usize + etext as usize) / 2).into();
    let mid_rodata: VirtAddr = ((srodata as usize + erodata as usize) / 2).into();
    let mid_data: VirtAddr = ((sdata as usize + edata as usize) / 2).into();
    assert!(!kernel_space
        .page_table
        .translate(mid_text.floor())
        .unwrap()
        .writable(),);
    assert!(!kernel_space
        .page_table
        .translate(mid_rodata.floor())
        .unwrap()
        .writable(),);
    assert!(!kernel_space
        .page_table
        .translate(mid_data.floor())
        .unwrap()
        .executable(),);
    println!("remap_test passed!");
}
