//! File and filesystem-related syscalls
use crate::fs::{is_abs_path, make_pipe, open_file, KStat, OSInode, OpenFlags};
use crate::mm::{translated_byte_buffer, translated_str, translated_refmut, UserBuffer};
use crate::task::{current_task, current_user_token};
use alloc::string::{String, ToString};
use alloc::vec::Vec;
use easy_fs::{EasyFileSystem, Inode};
use alloc::sync::Arc;

pub fn sys_dup(fd: usize) -> isize {
    let task = current_task().unwrap();
    let mut inner = task.inner_exclusive_access();
    let new_fd = inner.alloc_fd(0);
    inner.fd_table[new_fd] = inner.fd_table[fd].clone();
    print!("sys_dup(fd: {}) = {}", fd, new_fd);
    new_fd as isize
}

pub fn sys_dup3(old_fd: usize, new_fd: usize) -> isize {
    let task = current_task().unwrap();
    let mut inner = task.inner_exclusive_access();
    while new_fd >= inner.fd_table.len() {
        inner.fd_table.push(None);
    }
    inner.fd_table[new_fd] = inner.fd_table[old_fd].clone();
    print!(
        "sys_dup3(old_fd: {}, new_fd: {}) = {}",
        old_fd,
        new_fd,
        new_fd as isize
    );
    new_fd as isize
}


pub fn sys_write(fd: usize, buf: *const u8, len: usize) -> isize {
    let token = current_user_token();
    let task = current_task().unwrap();
    let inner = task.inner_exclusive_access();
    if fd >= inner.fd_table.len() {
        return -1;
    }
    if let Some(file) = &inner.fd_table[fd] {
        if !file.writable() {
            return -1;
        }
        let file = file.clone();
        // release current task TCB manually to avoid multi-borrow
        drop(inner);
        file.write(UserBuffer::new(translated_byte_buffer(token, buf, len))) as isize
    } else {
        -1
    }
}

pub fn sys_read(fd: usize, buf: *const u8, len: usize) -> isize {
    let token = current_user_token();
    let task = current_task().unwrap();
    let inner = task.inner_exclusive_access();
    if fd >= inner.fd_table.len() {
        return -1;
    }
    if let Some(file) = &inner.fd_table[fd] {
        let file = file.clone();
        if !file.readable() {
            return -1;
        }
        // release current task TCB manually to avoid multi-borrow
        drop(inner);
        file.read(UserBuffer::new(translated_byte_buffer(token, buf, len))) as isize
    } else {
        -1
    }
}

pub fn sys_open_at(path: *const u8, flags: u32, fd_i: isize, _mode: u32) -> isize {
    let task = current_task().unwrap();
    let token = current_user_token();
    let path = translated_str(token, path);
    
    if let Some(inode) = open_file(path.as_str(), OpenFlags::from_bits(flags).unwrap()) {
        let mut inner = task.inner_exclusive_access();
        let fd = inner.alloc_fd(0);
        inner.fd_table[fd] = Some(inode);
        // 
        if fd_i == fd as isize{
            
            fd as isize
        }else {
            print!("The input fd is error --- no find");
            -1
        }
    } else {
        -1
    }
}

pub fn sys_close(fd: usize) -> isize {
    let task = current_task().unwrap();
    let mut inner = task.inner_exclusive_access();
    if fd >= inner.fd_table.len() {
        return -1;
    }
    if inner.fd_table[fd].is_none() {
        return -1;
    }
    inner.fd_table[fd].take();
    0
}

pub fn sys_pipe2(pipe: *mut usize) -> isize {
    let task = current_task().unwrap();// 获取当前任务
    let token = current_user_token();// 获取当前用户令牌
    let mut inner = task.inner_exclusive_access();// 在任务的内部独占访问中获取可变引用
    let (pipe_read, pipe_write) = make_pipe(); // 使用 make_pipe() 函数创建管道，返回读取和写入端的文件描述符
    // 分配一个文件描述符用于读取端，并将其与 pipe_read 关联
    let read_fd = inner.alloc_fd(0);
    inner.fd_table[read_fd] = Some(pipe_read);
    
    // 分配一个文件描述符用于写入端，并将其与 pipe_write 关联
    let write_fd = inner.alloc_fd(0);
    inner.fd_table[write_fd] = Some(pipe_write);

    *translated_refmut(token, pipe) = read_fd;// 将读取端的文件描述符存储在 pipe 指向的位置上
    *translated_refmut(token, unsafe { pipe.add(1) }) = write_fd;// 将写入端的文件描述符存储在 pipe + 1 指向的位置上
    
    // 返回执行成功的标志（0）
    0
}

pub fn sys_fstat(_fd: isize, _buf: *mut u8) -> isize {
    // let token = current_user_token();
    // let task = current_task().unwrap();
    // let inner = task.inner_exclusive_access();
    // let buf_vec = translated_byte_buffer(token, buf, core::mem::size_of::<KStat>());
    // let userbuf = UserBuffer::new(buf_vec);

    // if fd < 0 || fd >= inner.fd_table.len() as isize {
    //     return -1;
    // }
    // let file_opt = inner.fd_table[fd as usize].clone();
    // drop(inner); // 手动释放锁，以避免多重引用

    // if let Some(file) = file_opt {
    //     // 这里假设 file 是 Arc<dyn fs::File + Send + Sync>
    //     let buf_vec = translated_byte_buffer(token, buf, core::mem::size_of::<KStat>());
    //     let mut userbuf = UserBuffer::new(buf_vec);

    //     // 获取文件的状态信息并填充到用户提供的缓冲区
    //     userbuf.write(file.stat().as_bytes());

    //     0
    // } else {
    //     // 处理 Option 中为 None 的情况
    //     -1
    // }
    0
}

// 解析路径，获取父目录路径和目录名
fn parse_parent_and_name(cwd: &str, path: &str) -> (String, String) {
    // 如果路径是绝对路径，则直接使用
    if path.starts_with('/') {
        let mut components = path.rsplitn(2, '/');
        let dir_name = components.next().unwrap_or("").to_string();
        let parent_path = components.next().unwrap_or("/").to_string();
        return (parent_path, dir_name);
    }

    // 否则，路径是相对路径，需要与当前工作目录拼接
    let mut components = cwd.split('/');
    let path_components = path.split('/');
    let mut combined_components: Vec<&str> = Vec::new();

    // 将当前工作目录的组件添加到组合组件中
    combined_components.extend(components.by_ref());

    // 将相对路径的组件添加到组合组件中
    combined_components.extend(path_components);

    // 获取目录名
    let dir_name = combined_components.last().unwrap_or(&"").to_string();

    // 获取父目录路径
// 获取父目录路径
    let parent_path = combined_components
        .iter()
        .take(combined_components.len() - 1)
        .map(|&s| s.to_string())  // 将 &str 转换为 String
        .collect::<Vec<_>>()
        .join("/");


    (parent_path, dir_name)
}


pub fn sys_mkdirat(dirfd: isize, path: *const u8, _mode: u32) -> isize {
    
    // 获取当前用户的令牌和任务
    let token = current_user_token();
    let task = current_task().unwrap();
    
    // 将路径转换为字符串
    let path_str = translated_str(token, path);

    // 获取当前工作目录
    let cwd = if dirfd == -100 && !is_abs_path(&path_str) {
        task.inner_exclusive_access().cwd.clone()
    } else {
        String::from("/")
    };

    // 解析路径，获取父目录和目录名
    let (parent_path, dir_name) = parse_parent_and_name(&cwd, &path_str);
    

    // // 获取父目录的 Inode
    // let parent_inode = match Inode::get_inode_by_path(&current_inode ,&parent_path) {
    //     Some(inode) => inode,
    //     None => {
    //         // 处理找不到父目录的情况
    //         print!("找不到父目录");
    //         return -1;  // 返回错误码
    //     }
    // };
    // // 在父目录下创建目录
    // match parent_inode.create_dir(&dir_name) {
    //     Some(new_inode) => {
    //         // 成功创建目录
    //         // 这里可以执行其他操作，比如设置权限等
            
    //         0 // 返回成功的标志
    //     }
    //     None => {
    //         // 创建目录失败
    //         print!("创建目录失败");
    //         -1 // 返回错误码
    //     }
    // }
    let result = match open_file(&dir_name, OpenFlags::O_RDONLY) {
        Some(inode) => {
            // 获取父目录的 Inode
            let parent_inode = match Inode::get_inode_by_path(&inode.get_inode(), &parent_path) {
                Some(inode) => inode,
                None => {
                    // 处理找不到父目录的情况
                    print!("找不到父目录");
                    return -1;  // 返回错误码
                }
            };
    
            // 在父目录下创建目录
            match parent_inode.create_dir(&dir_name) {
                Some(_new_inode) => {
                    // 成功创建目录
                    // 这里可以执行其他操作，比如设置权限等
                    0 // 返回成功的标志
                }
                None => {
                    // 创建目录失败
                    print!("创建目录失败");
                    -1 // 返回错误码
                }
            }
        }
        None => {
            // 未能打开文件的情况
            print!("无法打开文件");
            -1 // 返回错误码
        }
    };
    result
    
}
