use crate::config::PAGE_SIZE;
use crate::fs::{open_file, OpenFlags};
use crate::mm::{translated_refmut, translated_str,translate_writable_va, translated_byte_buffer, UserBuffer, translated_ref};
use crate::mm;
use crate::task::{
    add_task, current_task, current_user_token, exit_current_and_run_next,
    suspend_current_and_run_next, munmap, mmap, Utsname, CloneFlags,
};
use crate::timer::*;
use alloc::sync::Arc;
use alloc::vec::Vec;
use core::mem::size_of;


pub fn sys_exit(exit_code: i32) -> ! {
    exit_current_and_run_next(exit_code);
    panic!("Unreachable in sys_exit!");
}

pub fn sys_yield() -> isize {
    suspend_current_and_run_next()
}
pub fn sys_get_time_ms() -> isize {
    get_time_ms() as isize
}
///获取进程已执行的秒数
pub fn sys_times(time:  *mut i64) -> isize{
    let token = current_user_token();
    let sec = get_time_us() as i64;
    *translated_refmut(token, time) = sec;
    *translated_refmut(token, unsafe { time.add(1) }) = sec;
    *translated_refmut(token, unsafe { time.add(2) }) = sec;
    *translated_refmut(token, unsafe { time.add(3) }) = sec;
    0
}
pub fn sys_uname(buf: *mut u8) -> isize {
    // let uname = utsname {
    //     sysname: b"OurOS\0",
    //     nodename:  b"OurOS\0",
    //     release:  b"Alpha\0",
    //     version:  b"1.1\0",
    //     machine:  b"RISC-V64\0",
    //     domainname: b"TCoreTEAM/OurOS\0"
    // };
    let token = current_user_token();
    let buf_vec = translated_byte_buffer(token, buf, size_of::<Utsname>());
    let uname = Utsname::new();
    // 使用UserBuffer结构，以便于跨页读写
    let mut userbuf = UserBuffer::new(buf_vec);
    userbuf.write(uname.as_bytes());
    0
}

/// 获取系统时间并输出至提供的内存中
pub fn sys_get_time(ts: usize, tz: usize) -> isize {
    let token = current_user_token();// 获取当前用户的访问令牌
    let mut pas: Vec<*mut usize> = Vec::new();// 准备一个指针数组来存储物理地址
    // 将输出时间戳输出内存地址转换成物理地址并添加到指针数组中
    match mm::translate_writable_va(token, ts) {
        Err(_) => return -1,
        Ok(pa) => pas.push(pa as *mut usize),
    }
    // 计算下一个内存地址的位置，并将其地址转换成物理地址添加到指针数组中
    match mm::translate_writable_va(token, ts + size_of::<usize>()) {
        Err(_) => return -1,
        Ok(pa) => pas.push(pa as *mut usize),
    }
    get_time_with_timespec(pas, tz)// 将内存地址的物理地址作为参数传递
}
pub fn sys_getppid() -> isize{
    let current_task = current_task().unwrap();
    // find a parent process

    // ---- access current PCB exclusively
    let mut inner = current_task.inner_exclusive_access();
    if inner.parent.is_none()
    {
        return -1;
        // ---- release current PCB
    }
    else {
        
        let mut _pa = &mut inner.parent;
        if _pa.is_some() {
            if let Some(parent_weak) = _pa.clone() {
                if let Some(parent) = parent_weak.upgrade() {
                    *_pa = Some(Arc::downgrade(&parent));
                    println!("Parent process confirmed!------From SYS_getppid");
                    // ---- release current PCB
                    return parent.pid.0 as isize;
                } else {
                    println!("Parent process does not exist.");
                    // ---- release current PCB
                    return -1;
                }
            } else {
                println!("No parent process defined.");
                // ---- release current PCB
                return -1;
            }
        } else {
            // ---- release current PCB
            return -1;
        }
    }
}
pub fn sys_getpid() -> isize {
    current_task().unwrap().pid.0 as isize
}
#[allow(unused)]
pub fn sys_munmap1(start: usize, len: usize) -> isize {
    munmap(start, len).unwrap_or(-1)
}
pub fn sys_munmap(start: usize, len: usize) -> isize {
    let task = current_task().unwrap();
    let ret = task.munmap(start, len);
    ret
}
pub fn sys_fork(flags: usize, stack_ptr: usize, ptid: usize, ctid: usize, newtls: usize) -> isize {
    let current_task = current_task().unwrap();
    let flag = CloneFlags::from_bits(flags).unwrap();
    if !(flag.contains(CloneFlags::SIGCHLD)
        ||flag.contains(CloneFlags::CLONE_CHILD_SETTID)
        ||(flag.contains(CloneFlags::MANUAL_SELECT_PID))){
        return -1;
    }
    let new_task = current_task.fork(flags,ctid);
    let new_pid = new_task.pid.0;
    if stack_ptr != 0{
        let trap_cx = new_task.inner_exclusive_access().get_trap_cx();
        trap_cx.set_sp(stack_ptr);
    }
    // modify trap context of new_task, because it returns immediately after switching
    let trap_cx = new_task.inner_exclusive_access().get_trap_cx();
    // we do not have to move to next instruction since we have done it before
    // for child process, fork returns 0
    trap_cx.x[10] = 0;
    // add new task to scheduler
    add_task(new_task);
    println!("sys_fork(flags: {:?}, stack_ptr: 0x{:X}, ptid: {}, ctid: {}, newtls: {}) = {}", flags, stack_ptr, ptid, ctid, newtls, new_pid);
    new_pid as isize
}

pub fn sys_exec(path: *const u8) -> isize {
    let token = current_user_token();
    let path = translated_str(token, path);
    if let Some(app_inode) = open_file(path.as_str(), OpenFlags::O_RDONLY) {
        let all_data = app_inode.read_all();
        let task = current_task().unwrap();
        task.exec(all_data.as_slice());
        0
    } else {
        -1
    }
}
pub fn sys_sbrk(grow_size: isize, _is_shrink: usize) -> isize {
    let current_va = current_task().unwrap().grow_proc(grow_size) as isize;
    current_va
}

#[allow(unused)]
pub fn sys_brk(brk_addr: usize) -> isize{
    let mut addr_new = 0;
    if brk_addr == 0 {
        addr_new = sys_sbrk(0, 0) as usize;
    }
    else{
        let former_addr = current_task().unwrap().grow_proc(0);
        let grow_size: isize = (brk_addr - former_addr) as isize;
        addr_new = current_task().unwrap().grow_proc(grow_size);
    }
    
    addr_new as isize
}
#[allow(unused)]
pub fn sys_mmap1(start: usize, len: usize, port: usize) -> isize {
    mmap(start, len, port).unwrap_or(-1)
}
pub fn sys_mmap(start: usize, len: usize, prot: usize, flags: usize, fd: isize, off: usize) -> isize {
    let task = current_task().unwrap();
    let mut adjust_len = len;
    if adjust_len == 0{
        adjust_len = PAGE_SIZE;
    }
    let result_addr = task.mmap(start, adjust_len, prot, flags, fd, off);
    return result_addr as isize;
}
pub fn sys_sleep(time_req: *mut u64, time_remain: *mut u64) -> isize{
    let token = current_user_token();
    let sec = *translated_ref(token, time_req) as usize;
    let usec = *translated_ref(token, unsafe{time_req.add(1)}) as usize;
    let (end_sec,end_usec) = sum_time(sec, usec, get_time_s(), get_time_us());
    loop{
        let cur_sec = get_time_s();
        let cur_usec = get_time_us();
        if compare_time(end_sec, end_usec, cur_sec, cur_usec) {
            suspend_current_and_run_next(); 
        }
        else{
            if time_remain as usize != 0{
                *translated_refmut(token, time_remain) = 0 as u64;
            }
            *translated_refmut(token, unsafe { time_remain.add(1) }) = 0 as u64;
            return 0;
        }                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      
    }
}
/// If there is not a child process whose pid is same as given, return -1.
/// Else if there is a child process but it is still running, return -2.
pub fn sys_waitpid(pid: isize, exit_code_ptr: *mut i32) -> isize {
    let task = current_task().unwrap();
    // find a child process

    // ---- access current PCB exclusively
    let mut inner = task.inner_exclusive_access();
    if !inner
        .children
        .iter()
        .any(|p| pid == -1 || pid as usize == p.getpid())
    {
        return -1;
        // ---- release current PCB
    }
    let pair = inner.children.iter().enumerate().find(|(_, p)| {
        // ++++ temporarily access child PCB exclusively
        p.inner_exclusive_access().is_zombie() && (pid == -1 || pid as usize == p.getpid())
        // ++++ release child PCB
    });
    if let Some((idx, _)) = pair {
        let child = inner.children.remove(idx);
        // confirm that child will be deallocated after being removed from children list
        assert_eq!(Arc::strong_count(&child), 1);
        let found_pid = child.getpid();
        // ++++ temporarily access child PCB exclusively
        let exit_code = child.inner_exclusive_access().exit_code;
        // ++++ release child PCB
        *translated_refmut(inner.memory_set.token(), exit_code_ptr) = exit_code;
        found_pid as isize
    } else {
        -2
    }
    // ---- release current PCB automatically
}
pub fn sys_wait4(pid: isize, wstatus: *mut i32, option: isize) -> isize {
    if option != 0 {
        panic!{"Extended option not support yet..."};
    }

    loop {
        let task = current_task().unwrap();
        let mut inner = task.inner_exclusive_access();
        if inner.children
            .iter()
            .find(|p| {pid == -1 || pid as usize == p.getpid()})
            .is_none() {
            return -1;
            // ---- release current PCB lock
            }
        let waited = inner.children
            .iter()
            .enumerate()
            .find(|(_, p)| {
                // ++++ temporarily hold child PCB lock
                p.inner_exclusive_access().is_zombie() && (pid == -1 || pid as usize == p.getpid())
                // ++++ release child PCB lock
            });
        if let Some((idx,_)) = waited {
            let waited_child = inner.children.remove(idx);
            // confirm that child will be deallocated after being removed from children list
            // println!("[wait4]:pid {} child_pid {} ", task.pid.0, waited_child.getpid());
            assert_eq!(Arc::strong_count(&waited_child), 1);
            let found_pid = waited_child.getpid();
            // ++++ temporarily hold child lock
            let exit_code = waited_child.inner_exclusive_access().exit_code;
            let ret_status = exit_code << 8;
            if (wstatus as usize) != 0{
                *translated_refmut(inner.memory_set.token(), wstatus) = ret_status;
            }
            // println!("=============The pid being waited is {}===================", pid);
            // println!("=============The exit code of waiting_pid is {}===========", exit_code);
            return found_pid as isize;
        } else {
            drop(inner);
            drop(task);
            //print!("\n");
            //print!(" ");
            suspend_current_and_run_next();
            // continue;
        }
    }
    // ---- release current PCB automatically
}
