//!Implementation of [`TaskControlBlock`]
use super::{TaskContext, CloneFlags};
use super::{pid_alloc, KernelStack, PidHandle};
use crate::config::{TRAP_CONTEXT, USER_HEAP_SIZE, PAGE_SIZE, MMAP_BASE};
use crate::fs::{File, Stdin, Stdout, FileDescripter};
use crate::mm::{MemorySet, PhysPageNum, VirtAddr, KERNEL_SPACE, MapPermission, MmapArea};
use crate::sync::UPSafeCell;
use crate::trap::{trap_handler, TrapContext};
use alloc::sync::{Arc, Weak};
use alloc::vec;
use alloc::vec::Vec;
use core::cell::RefMut;
use spin::{Mutex, MutexGuard};
use alloc::string::String;

///任务控制块
pub struct TaskControlBlock {
    /// 进程的标识符（PID），用于唯一标识一个进程。
    pub pid: PidHandle,
    /// 进程的内核栈，用于存储进程在内核执行期间的状态。
    pub kernel_stack: KernelStack,
    /// 一个可变字段，包含一个 `UPSafeCell` 类型的引用，指向 `TaskControlBlockInner` 结构体。
    inner: UPSafeCell<TaskControlBlockInner>,
}
///文件描述符表
pub type FdTable =  Vec<Option<FileDescripter>>;
pub struct TaskControlBlockInner {
    pub trap_cx_ppn: PhysPageNum,
    pub base_size: usize,
    pub task_cx: TaskContext,
    pub task_status: TaskStatus,
    pub memory_set: MemorySet,
    pub heap_start: usize,
    pub heap_pt: usize,
    pub mmap_area: MmapArea,
    pub parent: Option<Weak<TaskControlBlock>>,
    pub children: Vec<Arc<TaskControlBlock>>,
    pub exit_code: i32,
    pub fd_table: Vec<Option<Arc<dyn File + Send + Sync>>>,
    pub cwd: String,
}

impl TaskControlBlockInner {
    pub fn get_trap_cx(&self) -> &'static mut TrapContext {
        self.trap_cx_ppn.get_mut()
    }
    pub fn get_user_token(&self) -> usize {
        self.memory_set.token()
    }
    fn get_status(&self) -> TaskStatus {
        self.task_status
    }
    pub fn is_zombie(&self) -> bool {
        self.get_status() == TaskStatus::Zombie
    }
    pub fn alloc_fd(&mut self, minfd: usize) -> usize {
        let len = self.fd_table.len();
        (minfd..len)
            .find(|&idx| self.fd_table[idx].is_none())
            .unwrap_or_else(|| {
                self.fd_table.push(None);
                len
            })
    }
    pub fn mmap(&mut self, start: usize, len: usize, port: usize) -> Result<isize, isize> {
        self.memory_set.mmap(start, len, port)
    }
    pub fn munmap(&mut self, start: usize, len: usize) -> Result<isize, isize> {
        self.memory_set.munmap(start, len)
    }

}

impl TaskControlBlock {
    ///获取对任务控制块内部状态的独占访问权限
    pub fn inner_exclusive_access(&self) -> RefMut<'_, TaskControlBlockInner> {
        self.inner.exclusive_access()
    }
    ///根据提供的 ELF 文件数据创建一个新的任务控制块
    pub fn new(elf_data: &[u8]) -> Self {
        // memory_set with elf program headers/trampoline/trap context/user stack
        let (memory_set, user_sp, user_heap, entry_point) = MemorySet::from_elf(elf_data);
        let trap_cx_ppn = memory_set
            .translate(VirtAddr::from(TRAP_CONTEXT).into())
            .unwrap()
            .ppn();
        // alloc a pid and a kernel stack in kernel space
        let pid_handle = pid_alloc(17,0);
        let kernel_stack = KernelStack::new(&pid_handle);
        let kernel_stack_top = kernel_stack.get_top();
        let task_control_block = Self {
            pid: pid_handle,
            kernel_stack,
            inner: unsafe {
                UPSafeCell::new(TaskControlBlockInner {
                    trap_cx_ppn,
                    base_size: user_sp,
                    heap_start: user_heap,
                    heap_pt: user_heap,
                    task_cx: TaskContext::goto_trap_return(kernel_stack_top),
                    task_status: TaskStatus::Ready,
                    memory_set,
                    mmap_area: MmapArea::new(VirtAddr::from(MMAP_BASE), VirtAddr::from(MMAP_BASE)),
                    parent: None,
                    children: Vec::new(),
                    exit_code: 0,
                    fd_table: vec![
                        // 0 -> stdin
                        Some(Arc::new(Stdin)),
                        // 1 -> stdout
                        Some(Arc::new(Stdout)),
                        // 2 -> stderr
                        Some(Arc::new(Stdout)),
                    ],
                    cwd: String::from("/"),
                })
            },
        };
        // prepare TrapContext in user space
        let trap_cx = task_control_block.inner_exclusive_access().get_trap_cx();
        *trap_cx = TrapContext::app_init_context(
            entry_point,
            user_sp,
            KERNEL_SPACE.exclusive_access().token(),
            kernel_stack_top,
            trap_handler as usize,
        );
        task_control_block
    }
    ///执行一个程序
    pub fn exec(&self, elf_data: &[u8]) {
        // memory_set with elf program headers/trampoline/trap context/user stack
        let (memory_set, user_sp, _user_heap, entry_point) = MemorySet::from_elf(elf_data);
        let trap_cx_ppn = memory_set
            .translate(VirtAddr::from(TRAP_CONTEXT).into())
            .unwrap()
            .ppn();

        // **** access current TCB exclusively
        let mut inner = self.inner_exclusive_access();
        // substitute memory_set
        inner.memory_set = memory_set;
        // update trap_cx ppn
        inner.trap_cx_ppn = trap_cx_ppn;
        // initialize trap_cx
        let trap_cx = TrapContext::app_init_context(
            entry_point,
            user_sp,
            KERNEL_SPACE.exclusive_access().token(),
            self.kernel_stack.get_top(),
            trap_handler as usize,
        );
        *inner.get_trap_cx() = trap_cx;
        // **** release current PCB
    }
    ///创建一个子进程
    pub fn fork(self: &Arc<TaskControlBlock>,flags: usize, ctid: usize) -> Arc<TaskControlBlock> {
        // ---- hold parent PCB lock
        let mut parent_inner = self.inner_exclusive_access();
        // copy user space(include trap context)
        let memory_set = MemorySet::from_existed_user(&parent_inner.memory_set);
        let trap_cx_ppn = memory_set
            .translate(VirtAddr::from(TRAP_CONTEXT).into())
            .unwrap()
            .ppn();                        
        let pid_handle = pid_alloc(flags,ctid);
        //aloc a kernel stack in kernel space
        let kernel_stack = KernelStack::new(&pid_handle);
        let kernel_stack_top = kernel_stack.get_top();
        // copy fd table
        let mut new_fd_table: Vec<Option<Arc<dyn File + Send + Sync>>> = Vec::new();
        for fd in parent_inner.fd_table.iter() {
            if let Some(file) = fd {
                new_fd_table.push(Some(file.clone()));
            } else {
                new_fd_table.push(None);
            }
        }
        let task_control_block = Arc::new(TaskControlBlock {
            pid: pid_handle,
            kernel_stack,
            inner: unsafe {
                UPSafeCell::new(TaskControlBlockInner {
                    trap_cx_ppn,
                    base_size: parent_inner.base_size,
                    heap_start: parent_inner.heap_start,
                    heap_pt: parent_inner.heap_pt,
                    task_cx: TaskContext::goto_trap_return(kernel_stack_top),
                    task_status: TaskStatus::Ready,
                    memory_set,
                    mmap_area: MmapArea::new(VirtAddr::from(MMAP_BASE), VirtAddr::from(MMAP_BASE)),
                    parent: Some(Arc::downgrade(self)),
                    children: Vec::new(),
                    exit_code: 0,
                    fd_table: new_fd_table,
                    cwd: parent_inner.cwd.clone(),
                })
            },
        });
        // add child
        parent_inner.children.push(task_control_block.clone());
        // modify kernel_sp in trap_cx
        // **** access child PCB exclusively
        let trap_cx = task_control_block.inner_exclusive_access().get_trap_cx();
        trap_cx.kernel_sp = kernel_stack_top;
        // return
        task_control_block
        // **** release child PCB
        // ---- release parent PCB
    }
    /// 获取当前进程的ID
    pub fn getpid(&self) -> usize {
        self.pid.0
    }
    /// 映射内存区域到进程地址空间
    pub fn mmap(&self, start: usize, len: usize, prot: usize, flags: usize, fd: isize, off: usize) -> usize {
        // gdb_println!(SYSCALL_ENABLE,"[mmap](0x{:X},{},{},0x{:X},{},{})",start, len, prot, flags, fd, off);
        
        if start % PAGE_SIZE != 0{
            panic!("mmap: start_va not aligned");
        } 
        let mut inner = self.inner_exclusive_access();
        let _fd_table = inner.fd_table.clone();
        let token = inner.get_user_token();
        let va_top = inner.mmap_area.get_mmap_top();
        let end_va = VirtAddr::from(va_top.0 + len);
        // "prot<<1" is equal to  meaning of "MapPermission"
        let map_flags = (((prot & 0b111)<<1) + (1<<4))  as u8; // "1<<4" means user
    
        let mut startvpn = start/PAGE_SIZE;
        
        if start != 0 { // "Start" va Already mapped
            while startvpn < (start+len)/PAGE_SIZE {
                if inner.memory_set.set_pte_flags(startvpn.into(), map_flags as usize) == -1{
                    panic!("mmap: start_va not mmaped");
                }
                startvpn +=1;
            }
            return start;
        }
        else{ // "Start" va not mapped
            //inner.memory_set.insert_kernel_mmap_area(va_top, end_va, MapPermission::from_bits(map_flags).unwrap());
            inner.memory_set.insert_mmap_area(va_top, end_va, MapPermission::from_bits(map_flags).unwrap());
            //inner.mmap_area.push_kernel(va_top.0, len, prot, flags, fd, off, fd_table, token);
            inner.mmap_area.push(va_top.0, len, prot, flags, fd, off, token);
            va_top.0
        }
    }
    ///取消映射内存区域到进程地址空间
    pub fn munmap(&self, start: usize, len: usize) -> isize {
        let mut inner = self.inner_exclusive_access();
        inner.memory_set.remove_area_with_start_vpn(VirtAddr::from(start).into());
        inner.mmap_area.remove(start, len)
    }
    /// 根据给定的 `grow_size` 参数，扩大或缩小进程堆的大小。
    pub fn grow_proc(&self, grow_size: isize) -> usize {
        let mut inner = self.inner_exclusive_access();
        if grow_size > 0 {
            let growed_addr: usize =inner.heap_pt + grow_size as usize;
            let limit = inner.heap_start + USER_HEAP_SIZE;
            if growed_addr > limit {
                panic!("process doesn't have enough memsize to grow! {} {} {} {}", limit, inner.heap_pt, growed_addr, self.pid.0);
            }
            inner.heap_pt = growed_addr;
        }
        else {
            let shrinked_addr: usize = inner.heap_pt + grow_size as usize;
            if shrinked_addr < inner.heap_start {
                panic!("Memory shrinked to the lowest boundary!")
            }
            inner.heap_pt = shrinked_addr;
        }
        return inner.heap_pt;
    }
}

#[derive(Copy, Clone, PartialEq)]
/// 表示任务的状态。
pub enum TaskStatus {
    /// 任务准备就绪，等待执行。
    Ready,
    /// 任务正在运行。
    Running,
    /// 任务处于僵尸状态，即已经运行完毕但尚未清理。
    Zombie,
}
