#![no_std]
#![feature(linkage)]
#![feature(panic_info_message)]
#![feature(alloc_error_handler)]

#[macro_use]
pub mod console;
mod lang_items;
mod syscall;

extern crate alloc;
#[macro_use]
extern crate bitflags;

use buddy_system_allocator::LockedHeap;
use syscall::*;

const USER_HEAP_SIZE: usize = 32768;

static mut HEAP_SPACE: [u8; USER_HEAP_SIZE] = [0; USER_HEAP_SIZE];

#[global_allocator]
static HEAP: LockedHeap = LockedHeap::empty();

#[alloc_error_handler]
pub fn handle_alloc_error(layout: core::alloc::Layout) -> ! {
    panic!("Heap allocation error, layout = {:?}", layout);
}

#[no_mangle]
#[link_section = ".text.entry"]
pub extern "C" fn _start() -> ! {
    unsafe {
        HEAP.lock()
            .init(HEAP_SPACE.as_ptr() as usize, USER_HEAP_SIZE);
    }
    exit(main());
}

#[linkage = "weak"]
#[no_mangle]
fn main() -> i32 {
    panic!("Cannot find main!");
}

bitflags! {
    pub struct OpenFlags: u32 {
        const O_RDONLY = 0;
        const O_WRONLY = 1 << 0;
        const O_RDWR = 1 << 1;
        const O_CREATE = 1 << 9;
        const O_TRUNC = 1 << 10;
    }
}

bitflags! {
    pub struct MOdeFlags: u32 {
        const RDONLY = 0;
        const WRONLY = 1 << 0;
        const RDWR = 1 << 1;
    }
}

pub fn dup(fd: usize) -> isize{
    sys_dup(fd)
}

pub fn dup3(old_fd: usize, new_fd: usize) -> isize {
    sys_dup3(old_fd,new_fd)
}

pub fn open(path: &str, flags: OpenFlags,mode:MOdeFlags,fd:i32) -> isize {
    sys_open_at(fd,path, flags.bits,mode.bits)
}
pub fn close(fd: usize) -> isize {
    sys_close(fd)
}
pub fn read(fd: usize, buf: &mut [u8]) -> isize {
    sys_read(fd, buf)
}
pub fn write(fd: usize, buf: &[u8]) -> isize {
    sys_write(fd, buf)
}
pub fn fstat(fd: usize,buf: &[u8]) -> isize{
    sys_fstat(fd, buf)
}
pub fn exit(exit_code: i32) -> ! {
    sys_exit(exit_code);
}
pub fn yield_() -> isize {
    sys_yield()
}
pub fn mmap(start: usize, len: usize, port: usize)-> isize {
    sys_mmap(start, len, port)
}
pub fn munmap(start: usize, len: usize) -> isize {
    sys_munmap(start,len)
}
///结构体：进程已执行的秒数
pub struct Tms {
    ///进程已执行的秒数
    pub sec: usize,
}

///Tms结构的实现
impl Tms {
    ///析构函数
    pub fn new() -> Self {
        Tms { sec: 0 }
    }
}
pub fn times() -> isize{
    let tms = Tms::new();
    match sys_times(&tms) {
        // 如果返回值为 0，表示成功获取到时间
        0 => {
            let milliseconds = tms.sec & 0xffff;// 将秒数取后四位（sec & 0xffff），并乘以 1000 得到毫秒数
            milliseconds as isize
        },
        _ => -1,// 如果返回值不为 0，表示获取时间失败
    }
}
#[repr(C)]
#[derive(Debug)]
pub struct Timespec {
    pub sec: usize,
    pub usec: usize,
}

impl Timespec {
    pub fn new() -> Self {
        Timespec { sec: 0, usec: 0 }
    }
}

pub fn get_time() -> isize{
    sys_get_time_ms()
}

/// 获取当前时间并返回以毫秒为单位的时间戳
pub fn get_time_of_day() -> isize {
    let ts = Timespec::new(); // 创建一个 Timespec 结构体来存储时间值
    // 调用 sys_get_time 函数来获取系统时间，并将时间保存在 ts 中
    match sys_get_time(&ts, 0) {
        // 如果返回值为 0，表示成功获取到时间
        0 => {
            let milliseconds = (ts.sec & 0xffff) * 1000;// 将秒数取后四位（sec & 0xffff），并乘以 1000 得到毫秒数
            let timestamp = milliseconds + ts.usec / 1000;// 将微秒数（ts.usec）除以 1000 得到毫秒数，并将其与前面的秒数相加
            timestamp as isize
        },
        _ => -1,// 如果返回值不为 0，表示获取时间失败
    }
}
pub fn getpid() -> isize {
    sys_getpid()
}
pub fn getppid() -> isize{
    sys_getppid()
}
pub fn fork_more(flags:usize,ctid:usize) -> isize {
    sys_fork(flags,ctid)
}
pub fn fork() -> isize {
    sys_fork(17,0)
}
pub fn exec(path: &str) -> isize {
    sys_exec(path)
}
pub fn wait(exit_code: &mut i32) -> isize {
    loop {
        match sys_waitpid(-1, exit_code as *mut _) {
            -2 => {
                yield_();
            }
            // -1 or a real pid
            exit_pid => return exit_pid,
        }
    }
}

pub fn waitpid(pid: usize, exit_code: &mut i32) -> isize {
    loop {
        match sys_waitpid(pid as isize, exit_code as *mut _) {
            -2 => {
                yield_();
            }
            // -1 or a real pid
            exit_pid => return exit_pid,
        }
    }
}
pub fn sleep(period_ms: usize) {
    let start = sys_get_time_ms();
    while sys_get_time_ms() < start + period_ms as isize {
        sys_yield();
    }
}
pub fn nanosleep(period_ms: usize) {
    sys_sleep(period_ms);
}
pub fn mkdir(dirfd: isize, path: &str, mode: u32){
    sys_mkdirat(dirfd,path,mode);
}