use crate::Timespec;
use crate::Tms;
use core::arch::asm;

const SYSCALL_DUP: usize = 23;
const SYSCALL_DUP3: usize = 24;
const SYSCALL_MKDIRAT: usize = 34;
const SYS_OPENAT: usize = 56;
const SYSCALL_CLOSE: usize = 57;
const SYSCALL_READ: usize = 63;
const SYSCALL_WRITE: usize = 64;
const SYSCALL_FSTAT: usize = 80;
const SYSCALL_EXIT: usize = 93;
const SYSCALL_NANOSLEEP: usize = 101;
const SYSCALL_SCHED_YIELD: usize = 124;
const SYSCALL_TIMES: usize = 153;
const SYSCALL_GETTIMEMS: usize = 168;
const SYSCALL_GETTIMEOFDAY: usize = 169;
const SYSCALL_GETPID: usize = 172;
const SYSCALL_GETPPID: usize = 173;
const SYSCALL_MUNMAP: usize = 215;
const SYSCALL_CLONE: usize = 220;
const SYSCALL_EXEC: usize = 221;
const SYSCALL_MMAP: usize = 222;
const SYSCALL_WAITPID: usize = 260;

fn syscall(id: usize, args: [usize; 6]) -> isize {
    let mut ret: isize;
    unsafe {
        asm!(
            "ecall",
            inlateout("x10") args[0] => ret,
            in("x11") args[1],
            in("x12") args[2],
            in("x13") args[3],
            in("x14") args[4],
            in("x15") args[5],
            in("x17") id
        );
    }
    ret
}

pub fn sys_dup(fd: usize) -> isize{
    syscall(SYSCALL_DUP, [fd as usize, 0, 0, 0, 0, 0])
}

pub fn sys_dup3(old_fd: usize, new_fd: usize) -> isize{
    syscall(SYSCALL_DUP3, [old_fd as usize,new_fd as usize, 0, 0, 0, 0])
}

pub fn sys_mkdirat(dirfd: isize, path: &str, mode: u32) -> isize {
    syscall(SYSCALL_MKDIRAT, [dirfd as usize,path.as_ptr() as usize,mode as usize,0,0])
}

pub fn sys_open_at(fd: i32,path: &str, flags: u32, mode: u32) -> isize {
    syscall(SYS_OPENAT, [path.as_ptr() as usize, flags as usize, fd as usize, mode as usize, 0, 0])
}
pub fn sys_close(fd: usize) -> isize {
    syscall(SYSCALL_CLOSE, [fd, 0, 0, 0, 0, 0])
}

pub fn sys_read(fd: usize, buffer: &mut [u8]) -> isize {
    syscall(
        SYSCALL_READ,
        [fd, buffer.as_mut_ptr() as usize, buffer.len(), 0, 0, 0],
    )
}

pub fn sys_write(fd: usize, buffer: &[u8]) -> isize {
    syscall(SYSCALL_WRITE, [fd, buffer.as_ptr() as usize, buffer.len(), 0, 0, 0])
}

pub fn sys_fstat(fd: usize,buffer: &[u8]) -> isize{
    syscall(SYSCALL_FSTAT, [fd,   buffer.as_ptr() as usize,0,0,0])
}

pub fn sys_exit(exit_code: i32) -> ! {
    syscall(SYSCALL_EXIT, [exit_code as usize, 0, 0, 0, 0, 0]);
    panic!("sys_exit never returns!");
}
pub fn sys_sleep(period_ms: usize) -> isize {
    let time = Timespec {
        sec: 0,
        usec: period_ms * 1000,
    };
    syscall(
        SYSCALL_NANOSLEEP,
        [
            &time as *const Timespec as usize,
            &time as *const Timespec as usize,
            0, 0, 0, 0
        ],
    )
}
pub fn sys_yield() -> isize {
    syscall(SYSCALL_SCHED_YIELD, [0, 0, 0, 0, 0, 0])
}
pub fn sys_mmap(start: usize, len: usize, port: usize) -> isize {
    syscall(SYSCALL_MMAP,[start, len, port, 0, 0, 0])
}
pub fn sys_munmap(start: usize, len: usize) -> isize {
    syscall(SYSCALL_MUNMAP,[start,len, 0, 0, 0, 0])
}
pub fn sys_get_time_ms() ->isize {
    syscall(SYSCALL_GETTIMEMS,[0, 0, 0, 0, 0, 0])
}

pub fn sys_times(tms: &Tms) -> isize {
    syscall(SYSCALL_TIMES,[tms as *const _ as usize, 0, 0, 0, 0, 0])
}
pub fn sys_get_time(ts: &Timespec, tz: usize) -> isize {
    syscall(SYSCALL_GETTIMEOFDAY, [ts as *const _ as usize, tz, 0, 0, 0, 0])
}

pub fn sys_getpid() -> isize {
    syscall(SYSCALL_GETPID, [0, 0, 0, 0, 0, 0])
}

pub fn sys_getppid() -> isize {
    syscall(SYSCALL_GETPPID, [0, 0, 0, 0, 0, 0])
}

pub fn sys_fork(flags: usize, ctid: usize) -> isize {
    syscall(SYSCALL_CLONE, [flags, 0, 0, ctid, 0, 0])
}
pub fn sys_exec(path: &str) -> isize {
    syscall(SYSCALL_EXEC, [path.as_ptr() as usize, 0, 0, 0, 0, 0])
}

pub fn sys_waitpid(pid: isize, exit_code: *mut i32) -> isize {
    syscall(SYSCALL_WAITPID, [pid as usize, exit_code as usize, 0, 0, 0, 0])
}
